#include "first.h"
#include "MenuTextEdit.h"

#include "Iff.h"
#include "Mouse.h"
#include "DInput.h"
#include "GL.h"
#include "Os.h"

#include <stdio.h>

const Tag TAG_TXEF = TAG(T,X,E,F); // text edit "high light" font
const Tag TAG_TXES = TAG(T,X,E,S); // text edit "high light" shared font
const Tag TAG_TXMX = TAG(T,X,M,X);
const Tag TAG_TXEM = TAG(T,X,E,M); // test edit minumum # of characters (before it accepts it)
const Tag TAG_ALLW = TAG(A,L,L,W);
const Tag TAG_DALW = TAG(D,A,L,W);
const Tag TAG_TXAV = TAG(T,X,A,V);
const Tag TAG_TXEV = TAG(T,X,E,V);
const Tag TAG_TXXP = TAG(T,X,X,P);

MenuTextEdit::MenuTextEdit(MenuElement *parent, int maxWidth, int maxHeight, Iff *iff)
: MenuText(parent, maxWidth, maxHeight, NULL)
{
	int i;

	init();
	// setup the default allowed array
	for (i = 0; i < MAX_CHAR; i++)
		characterAllowed[i] = isprint(i) ? true : false;

	// now allow special characters for foreign language version.
	// according to 'isprint' these are non-printable
	// (see also $/prophecy/dsrc/language/charset.txt)
	//
	// additions requested by translations department:
	//
	characterAllowed[128] = true; // 0x80   0x0080   #  LATIN CAPITAL LETTER C WITH CEDILLA
	characterAllowed[129] = true; // 0x81   0x0081   #  LATIN SMALL LETTER U WITH DIAERESIS
	characterAllowed[130] = true; // 0x82   0x0082   #  LATIN SMALL LETTER E WITH ACUTE
	characterAllowed[131] = true; // 0x83   0x0083   #  LATIN SMALL LETTER A WITH CIRCUMFLEX
	characterAllowed[132] = true; // 0x84   0x0084   #  LATIN SMALL LETTER A WITH DIAERESIS
	characterAllowed[133] = true; // 0x85   0x0085   #  LATIN SMALL LETTER A WITH GRAVE
	characterAllowed[135] = true; // 0x87   0x0087   #  LATIN SMALL LETTER C WITH CEDILLA
	characterAllowed[136] = true; // 0x88   0x0088   #  LATIN SMALL LETTER E WITH CIRCUMFLEX
	characterAllowed[137] = true; // 0x89   0x0089   #  LATIN SMALL LETTER E WITH DIAERESIS
	characterAllowed[138] = true; // 0x8A   0x008A   #  LATIN SMALL LETTER E WITH GRAVE
	characterAllowed[139] = true; // 0x8B   0x008B   #  LATIN SMALL LETTER I WITH DIAERESIS
	characterAllowed[140] = true; // 0x8C   0x008C   #  LATIN SMALL LETTER I WITH CIRCUMFLEX
	characterAllowed[142] = true; // 0x8E   0x008E   #  LATIN CAPITAL LETTER A WITH DIAERESIS
	characterAllowed[144] = true; // 0x90   0x0090   #  LATIN SMALL LETTER E WITH DIAERESIS
	characterAllowed[147] = true; // 0x93   0x0093   #  LATIN SMALL LETTER O WITH CIRCUMFLEX
	characterAllowed[148] = true; // 0x94   0x0094   #  LATIN SMALL LETTER O WITH DIAERESIS
	characterAllowed[150] = true; // 0x96   0x0096   #  LATIN SMALL LETTER U WITH CIRCUMFLEX
	characterAllowed[151] = true; // 0x97   0x0097   #  LATIN SMALL LETTER U WITH GRAVE
	characterAllowed[153] = true; // 0x99   0x0099   #  LATIN CAPITAL LETTER O WITH DIAERESIS
	characterAllowed[154] = true; // 0x9A   0x009A   #  LATIN CAPITAL LETTER U WITH DIAERESIS

	//
	// ISO 8859-1 & UNICODE compliant part of set:
	//
	characterAllowed[160] = true; // 0xA0   0x00A0   #   NO-BREAK SPACE
	characterAllowed[161] = true; // 0xA1   0x00A1   #   INVERTED EXCLAMATION MARK
	characterAllowed[162] = true; // 0xA2   0x00A2   #   CENT SIGN
	characterAllowed[163] = true; // 0xA3   0x00A3   #   POUND SIGN
	characterAllowed[164] = true; // 0xA4   0x00A4   #   CURRENCY SIGN
	characterAllowed[165] = true; // 0xA5   0x00A5   #   YEN SIGN
	characterAllowed[166] = true; // 0xA6   0x00A6   #   BROKEN BAR
	characterAllowed[167] = true; // 0xA7   0x00A7   #   SECTION SIGN
	characterAllowed[168] = true; // 0xA8   0x00A8   #   DIAERESIS
	characterAllowed[169] = true; // 0xA9   0x00A9   #   COPYRIGHT SIGN
	characterAllowed[170] = true; // 0xAA   0x00AA   #   FEMININE ORDINAL INDICATOR
	characterAllowed[171] = true; // 0xAB   0x00AB   #   LEFT-POINTING DOUBLE ANGLE QUOTATION MARK
	characterAllowed[172] = true; // 0xAC   0x00AC   #   NOT SIGN
	characterAllowed[173] = true; // 0xAD   0x00AD   #   SOFT HYPHEN
	characterAllowed[174] = true; // 0xAE   0x00AE   #   REGISTERED SIGN
	characterAllowed[175] = true; // 0xAF   0x00AF   #   MACRON
	characterAllowed[176] = true; // 0xB0   0x00B0   #   DEGREE SIGN
	characterAllowed[177] = true; // 0xB1   0x00B1   #   PLUS-MINUS SIGN
	characterAllowed[178] = true; // 0xB2   0x00B2   #   SUPERSCRIPT TWO
	characterAllowed[179] = true; // 0xB3   0x00B3   #   SUPERSCRIPT THREE
	characterAllowed[180] = true; // 0xB4   0x00B4   #   ACUTE ACCENT
	characterAllowed[181] = true; // 0xB5   0x00B5   #   MICRO SIGN
	characterAllowed[182] = true; // 0xB6   0x00B6   #   PILCROW SIGN
	characterAllowed[183] = true; // 0xB7   0x00B7   #   MIDDLE DOT
	characterAllowed[184] = true; // 0xB8   0x00B8   #   CEDILLA
	characterAllowed[185] = true; // 0xB9   0x00B9   #   SUPERSCRIPT ONE
	characterAllowed[186] = true; // 0xBA   0x00BA   #   MASCULINE ORDINAL INDICATOR
	characterAllowed[187] = true; // 0xBB   0x00BB   #   RIGHT-POINTING DOUBLE ANGLE QUOTATION MARK
	characterAllowed[188] = true; // 0xBC   0x00BC   #   VULGAR FRACTION ONE QUARTER
	characterAllowed[189] = true; // 0xBD   0x00BD   #   VULGAR FRACTION ONE HALF
	characterAllowed[190] = true; // 0xBE   0x00BE   #   VULGAR FRACTION THREE QUARTERS
	characterAllowed[191] = true; // 0xBF   0x00BF   #   INVERTED QUESTION MARK
	characterAllowed[192] = true; // 0xC0   0x00C0   #   LATIN CAPITAL LETTER A WITH GRAVE
	characterAllowed[193] = true; // 0xC1   0x00C1   #   LATIN CAPITAL LETTER A WITH ACUTE
	characterAllowed[194] = true; // 0xC2   0x00C2   #   LATIN CAPITAL LETTER A WITH CIRCUMFLEX
	characterAllowed[195] = true; // 0xC3   0x00C3   #   LATIN CAPITAL LETTER A WITH TILDE
	characterAllowed[196] = true; // 0xC4   0x00C4   #   LATIN CAPITAL LETTER A WITH DIAERESIS
	characterAllowed[197] = true; // 0xC5   0x00C5   #   LATIN CAPITAL LETTER A WITH RING ABOVE
	characterAllowed[198] = true; // 0xC6   0x00C6   #   LATIN CAPITAL LETTER AE
	characterAllowed[199] = true; // 0xC7   0x00C7   #   LATIN CAPITAL LETTER C WITH CEDILLA
	characterAllowed[200] = true; // 0xC8   0x00C8   #   LATIN CAPITAL LETTER E WITH GRAVE
	characterAllowed[201] = true; // 0xC9   0x00C9   #   LATIN CAPITAL LETTER E WITH ACUTE
	characterAllowed[202] = true; // 0xCA   0x00CA   #   LATIN CAPITAL LETTER E WITH CIRCUMFLEX
	characterAllowed[203] = true; // 0xCB   0x00CB   #   LATIN CAPITAL LETTER E WITH DIAERESIS
	characterAllowed[204] = true; // 0xCC   0x00CC   #   LATIN CAPITAL LETTER I WITH GRAVE
	characterAllowed[205] = true; // 0xCD   0x00CD   #   LATIN CAPITAL LETTER I WITH ACUTE
	characterAllowed[206] = true; // 0xCE   0x00CE   #   LATIN CAPITAL LETTER I WITH CIRCUMFLEX
	characterAllowed[207] = true; // 0xCF   0x00CF   #   LATIN CAPITAL LETTER I WITH DIAERESIS
	characterAllowed[208] = true; // 0xD0   0x00D0   #   LATIN CAPITAL LETTER ETH (Icelandic)
	characterAllowed[209] = true; // 0xD1   0x00D1   #   LATIN CAPITAL LETTER N WITH TILDE
	characterAllowed[210] = true; // 0xD2   0x00D2   #   LATIN CAPITAL LETTER O WITH GRAVE
	characterAllowed[211] = true; // 0xD3   0x00D3   #   LATIN CAPITAL LETTER O WITH ACUTE
	characterAllowed[212] = true; // 0xD4   0x00D4   #   LATIN CAPITAL LETTER O WITH CIRCUMFLEX
	characterAllowed[213] = true; // 0xD5   0x00D5   #   LATIN CAPITAL LETTER O WITH TILDE
	characterAllowed[214] = true; // 0xD6   0x00D6   #   LATIN CAPITAL LETTER O WITH DIAERESIS
	characterAllowed[215] = true; // 0xD7   0x00D7   #   MULTIPLICATION SIGN
	characterAllowed[216] = true; // 0xD8   0x00D8   #   LATIN CAPITAL LETTER O WITH STROKE
	characterAllowed[217] = true; // 0xD9   0x00D9   #   LATIN CAPITAL LETTER U WITH GRAVE
	characterAllowed[218] = true; // 0xDA   0x00DA   #   LATIN CAPITAL LETTER U WITH ACUTE
	characterAllowed[219] = true; // 0xDB   0x00DB   #   LATIN CAPITAL LETTER U WITH CIRCUMFLEX
	characterAllowed[220] = true; // 0xDC   0x00DC   #   LATIN CAPITAL LETTER U WITH DIAERESIS
	characterAllowed[221] = true; // 0xDD   0x00DD   #   LATIN CAPITAL LETTER Y WITH ACUTE
	characterAllowed[222] = true; // 0xDE   0x00DE   #   LATIN CAPITAL LETTER THORN (Icelandic)
	characterAllowed[223] = true; // 0xDF   0x00DF   #   LATIN SMALL LETTER SHARP S (German)
	characterAllowed[224] = true; // 0xE0   0x00E0   #   LATIN SMALL LETTER A WITH GRAVE
	characterAllowed[225] = true; // 0xE1   0x00E1   #   LATIN SMALL LETTER A WITH ACUTE
	characterAllowed[226] = true; // 0xE2   0x00E2   #   LATIN SMALL LETTER A WITH CIRCUMFLEX
	characterAllowed[227] = true; // 0xE3   0x00E3   #   LATIN SMALL LETTER A WITH TILDE
	characterAllowed[228] = true; // 0xE4   0x00E4   #   LATIN SMALL LETTER A WITH DIAERESIS
	characterAllowed[229] = true; // 0xE5   0x00E5   #   LATIN SMALL LETTER A WITH RING ABOVE
	characterAllowed[230] = true; // 0xE6   0x00E6   #   LATIN SMALL LETTER AE
	characterAllowed[231] = true; // 0xE7   0x00E7   #   LATIN SMALL LETTER C WITH CEDILLA
	characterAllowed[232] = true; // 0xE8   0x00E8   #   LATIN SMALL LETTER E WITH GRAVE
	characterAllowed[233] = true; // 0xE9   0x00E9   #   LATIN SMALL LETTER E WITH ACUTE
	characterAllowed[234] = true; // 0xEA   0x00EA   #   LATIN SMALL LETTER E WITH CIRCUMFLEX
	characterAllowed[235] = true; // 0xEB   0x00EB   #   LATIN SMALL LETTER E WITH DIAERESIS
	characterAllowed[236] = true; // 0xEC   0x00EC   #   LATIN SMALL LETTER I WITH GRAVE
	characterAllowed[237] = true; // 0xED   0x00ED   #   LATIN SMALL LETTER I WITH ACUTE
	characterAllowed[238] = true; // 0xEE   0x00EE   #   LATIN SMALL LETTER I WITH CIRCUMFLEX
	characterAllowed[239] = true; // 0xEF   0x00EF   #   LATIN SMALL LETTER I WITH DIAERESIS
	characterAllowed[240] = true; // 0xF0   0x00F0   #   LATIN SMALL LETTER ETH (Icelandic)
	characterAllowed[241] = true; // 0xF1   0x00F1   #   LATIN SMALL LETTER N WITH TILDE
	characterAllowed[242] = true; // 0xF2   0x00F2   #   LATIN SMALL LETTER O WITH GRAVE
	characterAllowed[243] = true; // 0xF3   0x00F3   #   LATIN SMALL LETTER O WITH ACUTE
	characterAllowed[244] = true; // 0xF4   0x00F4   #   LATIN SMALL LETTER O WITH CIRCUMFLEX
	characterAllowed[245] = true; // 0xF5   0x00F5   #   LATIN SMALL LETTER O WITH TILDE
	characterAllowed[246] = true; // 0xF6   0x00F6   #   LATIN SMALL LETTER O WITH DIAERESIS
	characterAllowed[247] = true; // 0xF7   0x00F7   #   DIVISION SIGN
	characterAllowed[248] = true; // 0xF8   0x00F8   #   LATIN SMALL LETTER O WITH STROKE
	characterAllowed[249] = true; // 0xF9   0x00F9   #   LATIN SMALL LETTER U WITH GRAVE
	characterAllowed[250] = true; // 0xFA   0x00FA   #   LATIN SMALL LETTER U WITH ACUTE
	characterAllowed[251] = true; // 0xFB   0x00FB   #   LATIN SMALL LETTER U WITH CIRCUMFLEX
	characterAllowed[252] = true; // 0xFC   0x00FC   #   LATIN SMALL LETTER U WITH DIAERESIS
	characterAllowed[253] = true; // 0xFD   0x00FD   #   LATIN SMALL LETTER Y WITH ACUTE
	characterAllowed[254] = true; // 0xFE   0x00FE   #   LATIN SMALL LETTER THORN (Icelandic)
	characterAllowed[255] = true; // 0xFF   0x00FF   #   LATIN SMALL LETTER Y WITH DIAERESIS

	load(iff);
}

// ----------------------------------------------------------------------

MenuTextEdit::~MenuTextEdit(void)
{
	init();
}

// ----------------------------------------------------------------------

void MenuTextEdit::init(void)
{
	drawCursor                    = true;
	editFont                      = NULL;
	maxLength                     = 0;
	enterEvent                    = 0;
	escapeEvent                   = 0;
	isEditable                    = false;
	minimumNumChars               = 0;
	oldCapturer                   = NULL;
	deleteStringWhenCaptureStarts = false;
}

// ----------------------------------------------------------------------

void MenuTextEdit::load(Iff *iff)
{
	char *s;
	char  buffer[128];
	int   i, r, g, b;

	if (iff)
	{
		MenuText::load(iff);

		// by default, MenuTextEdit and MenuButtons are tab-stops.  Unless this is forced to be OFF
		if (tabStop != TabForceOn && tabStop != TabForceOff)
			tabStop = TabOn;

		// get the font color (if specified - defaults to MenuText::font)
		editFont = NULL;
		if (iff->seekChunk(TAG_TXEF, Iff::FlagOptional))
		{
			r = iff->read_int32();
			g = iff->read_int32();
			b = iff->read_int32();
			iff->read_string(buffer, sizeof(buffer));

			editFont = NEW Font(buffer,r,g,b);

			iff->exitChunk();
		}
		else
		{
			if (iff->seekChunk(TAG_TXES, Iff::FlagOptional))
			{
				iff->read_string(buffer, sizeof(buffer));
				editFont = findSharedFont(buffer);
				DEBUG_FATAL(!editFont, ("MENU_TEXT_EDIT cannot find shared font '%s'", buffer));
				iff->exitChunk();
			}
		}

		// load in the max length of the string
		maxLength = string.getLength();
		if (iff->seekChunk(TAG_TXMX, Iff::FlagOptional))
		{
			maxLength = iff->read_int32();
			iff->exitChunk();
		}

		// the event number to send when enter is presed
		if (iff->seekChunk(TAG_TXEV, Iff::FlagOptional))
		{
			enterEvent = iff->read_int32();
			iff->exitChunk();
		}																														  

		// the event number to send when esc is presed
		if (iff->seekChunk(TAG_TXAV, Iff::FlagOptional))
		{
			escapeEvent = iff->read_int32();
			iff->exitChunk();
		}

		// find out if there is to be a minimum number of characters in the string
		if (iff->seekChunk(TAG_TXEM, Iff::FlagOptional))
		{
			minimumNumChars = iff->read_int32();
			iff->exitChunk(); // 'TXEM'
		}

		// support additional characters
		if (iff->seekChunk(TAG_ALLW, Iff::FlagOptional))
		{
			s = iff->read_char(Iff::ReadRest);
			for (i = 0; s[i]; i++)
				characterAllowed[s[i]] = true;

			delete [] s;
			iff->exitChunk(); // 'ALLW'
		}

		// forbid additional characters
		if (iff->seekChunk(TAG_DALW, Iff::FlagOptional))
		{
			char *s;

			s = iff->read_char(Iff::ReadRest);
			for (i = 0; s[i]; i++)
				characterAllowed[s[i]] = false;

			delete [] s;
			iff->exitChunk(); // 'DALW'
		}

		// explicitly allowed characters ONLY
		if (iff->seekChunk(TAG_TXXP, Iff::FlagOptional))
		{
			char *s;

			for (i = 0; i<MAX_CHAR; i++)
				characterAllowed[i]=false;

			s = iff->read_char(Iff::ReadRest);
			for (i = 0; s[i]; i++)
				characterAllowed[s[i]] = true;

			delete [] s;
			iff->exitChunk(); // 'TXXP'
		}

		cursorTimer.setHz(4);
	}
}

// ----------------------------------------------------------------------

IoResult MenuTextEdit::event(IoEventStruct *event)
{
	real t;

	if (!isEditable && menuRect.isInInclusive(event->mouseX, event->mouseY))
	{
		// a left mouse button press in our rectangle will put us in edit mode
		if (!isEditable && event->type == IoEventMousePress && event->button == 0)
			startCapturingButPreserveOldStringInCaseTheUserHitsEscape();
	}
	else
		if (isEditable)
		{
			if (event->type == IoEventMouseSetCursor)
			{
				Mouse::setBitmap(NULL, 0);

				// update the blinking cursor
				t  = Clock::frameTime();
				if (isEditable)
					while (cursorTimer.updateSub(t))
					{
						t          = CONST_REAL(0);
						drawCursor = !drawCursor;
					}

				return IoResultBlock;
			}

			// if the user hits the delete key, then it will erase the line
			if (event->type == IoEventKeyPress && event->ch == DIK_DELETE)
				string.setNull();

			if (event->type == IoEventKeyChar)
			{
				switch (event->ch)
				{
					case '\b':
						if (string.getLength())
							string.setNull(string.getLength()-1);
						break;

					default:
						if (event->ch > 0 && event->ch < MAX_CHAR && characterAllowed[event->ch] && string.getLength()<maxLength)
							string.append(event->ch);
						break;
				}
			}

			if (event->type == IoEventKeyPress)
			{
				switch (event->ch)
				{
					case DIK_NUMPADENTER:
					case DIK_RETURN:
						// meet string qualifications
						if (static_cast<int>(string.getLength()) >= minimumNumChars)
						{
							if (enterEvent)
								menuEvent(enterEvent, this);
							stopCapturing();
						}
						break;

					case DIK_ESCAPE:
						if (preservedString.getLength())
							string = preservedString;
						if (escapeEvent)
							menuEvent(escapeEvent, this);
						stopCapturing();
						break;
				}
			}

			return IoResultBlock;
		}

	return IoResultPass;
}

// ----------------------------------------------------------------------

void MenuTextEdit::menuDraw(int x, int y) const
{
	Font *font;
	Rect r;
	int  rWd, rHt; // rectangle wd, ht
	int  sWd, sHt; // string wd, ht
	int  dx, dy;   // draw string at dx, dy

	MenuText::menuDraw(x, y);

	if (isEditable && drawCursor)
	{
		font =  getFont();
		r    =  getTextOutputRect();
		r.x0 += x;
		r.y0 += y;
		r.x1 += x;
		r.y1 += y;

		rWd  = r.x1 - r.x0;
		rHt  = r.y1 - r.y0;
		sWd  = font->pixelWidthString(getTextString());
		sHt  = font->pixelHeight();

		// figure out x pos (default to left justify)
		dx = r.x0 + sWd;
		if (justifyRule & Font::JR_right)
			dx = r.x1;
		if (justifyRule & Font::JR_centerX)
			dx = r.x1 - (rWd - sWd) / 2;
		// top justify
		dy = r.y0 - 1;

		Gl::setMaterial(MaterialList::white());
		Gl::line(    dx, dy,     dx, dy + sHt);
		Gl::line(dx + 1, dy, dx + 1, dy + sHt);
	}
}

// ----------------------------------------------------------------------

void MenuTextEdit::setTextString(const String &newText)
{
	string = newText;
}

// ----------------------------------------------------------------------

void MenuTextEdit::setTextString(int newTextInt)
{
	char text[64];

	sprintf(text, "%d", newTextInt);
	string.setNull();
	string.append(text,1);
}

// ----------------------------------------------------------------------

void MenuTextEdit::eraseText(void)
{
	string.setNull();
}

// ----------------------------------------------------------------------

Font *MenuTextEdit::getFont(void) const
{
	return isEditable && editFont ? editFont : font;
}

// ----------------------------------------------------------------------

void MenuTextEdit::startCapturing(void)
{
	// if people call this then eliminate the old (and current) string entirely
	preservedString.setNull();
	oldCapturer = IoWinManager::capture(this);
	if (deleteStringWhenCaptureStarts)
		string.setNull();
	isEditable = true;
	Os::pushRedirectionToMouse(0);
}

// ----------------------------------------------------------------------

void MenuTextEdit::startCapturingButPreserveOldStringInCaseTheUserHitsEscape(void)
{
	// if people call this then eliminate the old (and current) string entirely
	preservedString = string;
	oldCapturer = IoWinManager::capture(this);
	if (deleteStringWhenCaptureStarts)
		string.setNull();
	isEditable = true;
	Os::pushRedirectionToMouse(0);
}

// ----------------------------------------------------------------------

void MenuTextEdit::stopCapturing(void)
{
	isEditable = false;
	IoWinManager::release(this, oldCapturer);
	Os::popRedirectionToMouse();
}

// ----------------------------------------------------------------------
// this will allow for defaulting the "text" to a prompt: ie: "Click here to edit" and when the user
//  clicks on the editbox it will immediatly wax the string
void MenuTextEdit::setDeleteStringWhenCaptureStarts(bool deleteString)
{
	deleteStringWhenCaptureStarts = deleteString;
}

// ======================================================================
